import uuid
from itertools import chain, groupby

from django.db import connections, models

from arkindex.project.fields import Unnest


class ElementQuerySet(models.QuerySet):

    def trash(self, delete_children=True):
        """Performant deletion of any element queryset"""
        from arkindex.documents.models import Element, MetaData, Transcription, TranscriptionEntity, Classification, Selection
        from arkindex.dataimport.models import DataImport, DataImportElement

        # We only need the Element IDs to join on other models
        # Disable all ordering as it will not be required for deletions
        ids = self.order_by().values_list('id')

        # Generate the initial raw SQL query
        sql, params = ids.query.sql_with_params()

        # Delete all children from the targeted elements
        # using the same trash() method, iterating on every depth
        if delete_children is True:
            # First we need to find out the depths of the top elements' children
            # The fastest way is to calc all the lengths of children paths
            # by including the target SQL query, and joining it directly with paths
            # It's not possible to do that join with Django ORM
            with connections['default'].cursor() as cursor:
                cursor.execute("""select min(length), max(length) FROM (
                        select array_length(p.path, 1) as length
                        from documents_elementpath as p
                        inner join
                        ({}) as input on (array[input.id] && p.path)
                ) as lengths
                """.format(sql), params)
                min_paths, max_paths = cursor.fetchone()

            # Postgres will give us None when no children is found
            if min_paths is not None and max_paths is not None:
                # Django increments the path number to match the postgresl 1-index
                # Remove the lower levels first, then work the way up
                for i in reversed(range(min_paths - 1, max_paths)):
                    filters = {
                        f"paths__path__{i}__in": self,
                    }
                    Element.objects.filter(**filters).trash(delete_children=False)

        # TranscriptionEntity is linked through a transcription, so it's a bit harder to join
        TranscriptionEntity.objects.filter(transcription__element_id__in=ids).prefetch_related('transcription').delete()

        # We need to use raw deletion instead of the smarter delete from Django
        # to avoid loading in RAM the transcription ids to check that transcription entity is deleted cleanly
        Transcription.objects.filter(element_id__in=ids)._raw_delete(using='default')

        # Simple direct dependencies to remove, through a Foreign Key
        for cls in (MetaData, Classification, Selection, DataImportElement):
            cls.objects.filter(element_id__in=ids).delete()

        # Set elements directly on dataimports to None
        DataImport.objects.filter(element_id__in=ids).update(element_id=None)

        # This is where it gets really ugly: in order to delete all the children
        # without losing their reference, we need to inject the generated SQL query
        # directly into an SQL DELETE statement for paths
        # Once paths are deleted, we can finally delete the targeted elements
        with connections['default'].cursor() as cursor:
            cursor.execute("""
            WITH element_ids (id) AS (
                DELETE FROM documents_elementpath
                WHERE element_id IN ({})
                RETURNING element_id
            )
            DELETE FROM documents_element element
            USING element_ids
            WHERE element.id = element_ids.id
            """.format(sql), params)

        # Finally, delete top elements
        self.delete()


class ElementManager(models.Manager):
    """Model manager for elements"""

    def get_queryset(self):
        return ElementQuerySet(self.model, using=self._db)

    def get_ascending(self, child_id, recursive=True):
        """
        Get all parent elements for a specific element ID.
        > All the elements from paths
        """
        from arkindex.documents.models import ElementPath

        paths = ElementPath.objects.filter(element_id=child_id)
        if recursive:
            parent_ids = paths.annotate(parent_id=Unnest('path')).values('parent_id')
        else:
            parent_ids = paths.values('path__last')

        return self.filter(id__in=parent_ids)

    def get_ascendings_paths(self, *children_ids, **filters):
        """
        Get all ascending paths for some elements IDs.
        """
        from arkindex.documents.models import ElementPath
        # Load all parents
        parents = {
            parent.id: parent
            for parent in self.filter(
                **filters,
                id__in=ElementPath
                .objects
                .filter(element_id__in=children_ids)
                .annotate(parent_id=Unnest('path'))
                .values('parent_id')
            )
        }

        # Loads paths and group them by element ids
        paths = ElementPath.objects.filter(element_id__in=children_ids).order_by('element_id')
        tree = {
            elt_id: [p.path for p in elt_paths]
            for elt_id, elt_paths in groupby(paths, lambda e: e.element_id)
        }

        # Put Element instances in paths
        return {
            elt_id: [
                [
                    parents[parent_id]
                    for parent_id in path
                ]
                for path in paths
            ]
            for elt_id, paths in tree.items()
        }

    def get_descending(self, parent_id, **filters):
        """
        Get all child elements for a specific element ID.
        > All the elements where parent is in the path
        """
        return self.filter(paths__path__contains=[parent_id], **filters).order_by('paths__ordering').distinct()

    def get_descendings(self, parents_ids, prefetch=(), **filters):
        """
        Get all child elements for some element IDs.
        """
        assert all(isinstance(x, uuid.UUID) for x in parents_ids), 'UUIDs are required'

        # Loads paths and group them by element ids
        # The map to str is needed to avoid a cast error in PG
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(path__overlap=map(str, parents_ids)).order_by('element_id', 'ordering')
        tree = {
            parent_id: [
                p.element_id
                for p in paths
                if parent_id in p.path
            ]
            for parent_id in parents_ids
        }

        # Load children elements in bulk
        children = self.filter(id__in=[p.element_id for p in paths], **filters)
        if prefetch:
            children = children.prefetch_related(*prefetch)
        return {
            parent_id: [
                child
                for child in children
                if child.id in tree.get(parent_id)
            ]
            for parent_id in parents_ids
        }

    def get_neighbor_paths(self, element, n=1):
        """
        Get n neighbor paths: current, previous and next element paths

        This uses a single raw SQL query to use row numbers, instead of the ordering numeric values,
        to avoid issues with duplicated or missing orderings.

        * The `neighbors` subquery lists all paths for an element, and then list all possible neighbors
          with the same parents. An ordering consistent with rows numbers is added as `position`
        * The main query combines the `neighbors` subquery with itself to have a full product, and only retains
          the combinations with a distance under the requested limit, with elements of the same type
        * The output is ordered by parent path, then by position (row numbers)
        """
        from arkindex.documents.models import ElementPath
        return ElementPath.objects.raw(
            """
            with neighbors as (
              select link.*, row_number() over (partition by p.id order by link.ordering) as position
              from documents_elementpath p
              inner join documents_elementpath link on (p.path=link.path)
              inner join documents_element as e on (e.id=link.element_id and e.type_id = %(element_type)s)
              where p.element_id = %(element_id)s
            )

            select distinct n.*
            from neighbors as n
            inner join neighbors as m on (m.element_id = %(element_id)s and m.path = n.path)
            where abs(n.position - m.position) <= %(n)s
            order by n.path, n.position, n.element_id
            """,
            params={
                'element_id': element.id,
                'element_type': element.type_id,
                'n': n,
            }
        )

    def get_neighbors(self, element, n=1):
        """
        Returns a list of neighboring ElementPaths for an element, with a prefetched `element` attribute
        and a list of prefetched parent elements in the `parents` attribute.
        """
        paths = list(self.get_neighbor_paths(element, n))
        models.prefetch_related_objects(paths, 'element__type')
        # Build a set of all IDs to load from all paths, then load them into a dict
        elements = {
            elt.id: elt
            for elt in self.filter(id__in=set(chain(*(p.path for p in paths)))).select_related('type')
        }
        # Dispatch the elements by ID to each path
        for path in paths:
            path.parents = list(map(elements.get, path.path))
        return paths


class CorpusManager(models.Manager):
    '''
    Add ACL functions to corpus listing
    '''

    def readable(self, user, add_stats=False):
        qs = super().get_queryset().order_by('name')

        # add_stats is used to determine whether to return the count of users with rights on the corpora
        # concerned by the request. The count of users is made here because the application of certain
        # permissions made in the continuation of this function restricts the access to corpus rights which
        # does not concern the user if he is not an administrator
        if add_stats:
            qs = qs.annotate(authorized_users=models.Count('corpus_right'))

        # An anonymous has access to public corpus only
        if user.is_anonymous:
            return qs.filter(public=True)

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # An authenticated user has access to:
        # * public corpus
        # * private with ACL
        qs = qs.prefetch_related('corpus_right')
        qs = qs.filter(models.Q(public=True) | models.Q(corpus_right__user=user)).distinct()
        return qs

    def writable(self, user):
        # An anonymous user cannot write anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can write only on corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_write=True).distinct()

    def admin(self, user):
        # An anonymous user cannot manage anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can manage only corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_admin=True).distinct()
