from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureTestCase


class TestAdminAPI(FixtureTestCase):

    def test_reindex_requires_login(self):
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reindex_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reindex_requires_index_types(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'transcriptions': False,
            'elements': False,
            'entities': False,
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['At least one index type is required.'],
        })

    def test_reindex_drop_only_all(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'corpus': str(self.corpus.id),
            'drop': True,
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['`drop` can only be used when reindexing everything.'],
        })

    def test_reindex_element_in_corpus(self):
        corpus2 = Corpus.objects.create(name='some corpus')
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'corpus': str(corpus2.id),
            'element': str(self.corpus.elements.first().id),
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['The selected element is not in the selected corpus.'],
        })

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_reindex(self, delay_mock):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            element_id=None,
            corpus_id=None,
            entity_id=None,
            transcriptions=True,
            elements=True,
            entities=True,
            drop=False,
        ))

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_reindex_no_search(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(
            response.json(),
            ['Reindexation is unavailable due to the search feature being disabled.'],
        )
