from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree


class TestDestroyElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def test_element_destroy_verified_user(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_destroy_acl(self):
        self.private_corpus.corpus_right.create(user_id=self.user.id, can_write=True, can_admin=False)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        with self.assertNumQueries(4):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this element.'}
        )

    def test_element_destroy(self):
        self.client.force_login(self.user)
        castle_story = self.corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        self.assertEqual(self.corpus.elements.filter(id=castle_story.id).exists(), True)
        with self.assertExactQueries(
                'element_deletion.sql',
                params={
                    'id': str(castle_story.id),
                    'corpus_id': str(self.corpus.id),
                    'user_id': self.user.id
                }, skip=1):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(self.corpus.elements.filter(id=castle_story.id).exists(), False)

    def test_non_empty_element(self):
        """
        We can now delete a non-empty element
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

    def test_element_trash_children(self):
        # Check this volume has children
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 5)

        with self.assertExactQueries('element_trash_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())
        # Excluding the children, only the volume should have been deleted
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 4)

    def test_element_trash_without_children(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries('element_trash_no_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).trash(delete_children=False)

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        # No children were deleted
        self.assertEqual(Element.objects.filter(id__in=ids).count(), 24)

    def test_element_trash_no_order_by(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries('element_trash_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).order_by('name').trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())

    def test_element_trash_related_filter_bug(self):
        """
        This test should fail when a bug in ElementQuerySet.trash is fixed:
        Filtering by metadata, classification, selection, DataImportElement, or DataImport
        can cause elements to not actually be deleted as the related tables are deleted first.
        For example, deleting all elements with classifications will cause all classifications
        to be deleted, but not elements, because the QuerySet is re-evaluated
        and no elements will have classifications.
        """
        ml_class = self.corpus.ml_classes.create(name='Krabby')
        page1r, page1v, page2r = Element.objects.get_descending(self.vol.id).filter(type__slug='page')
        page1r.classifications.create(ml_class=ml_class, confidence=0.42)
        page1v.classifications.create(ml_class=ml_class, confidence=0.42)
        page2r.classifications.create(ml_class=ml_class, confidence=0.42)

        queryset = Element.objects.filter(classifications__isnull=False)

        # Ensure the pages are all in this queryset, and therefore should be deleted
        self.assertEqual(queryset.filter(id__in=[page1r.id, page1v.id, page2r.id]).count(), 3)

        with self.assertExactQueries('element_trash_ml_class.sql'):
            Element.objects.filter(classifications__isnull=False).trash()

        # Pages did not get deleted
        page1r.refresh_from_db()
        page1v.refresh_from_db()
        page2r.refresh_from_db()

        # Classifications have been deleted
        self.assertFalse(ml_class.classifications.exists())

    def test_element_trash_deep(self):
        """
        Tests the repetition of path deletions with deep hierarchies
        Builds a tree of elements A→B→C→D→E and deletes A with its children
        """
        elements = build_tree(
            {
                'B': 'A',
                'C': 'B',
                'D': 'C',
                'E': 'D',
            },
            type=self.corpus.types.create(slug='houndour', display_name='Houndour', folder=True),
            corpus=self.corpus,
        )

        with self.assertExactQueries('element_trash_deep.sql', params={'id': str(elements['A'].id)}):
            Element.objects.filter(id=elements['A'].id).trash()

        self.assertFalse(Element.objects.filter(id__in=[e.id for e in elements.values()]).exists())
